-------------------------------------------------------------------------------
--                G  O  B  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

-- Copied from Blizzard_TradeSkillRecipeList.lua
local LEARNED_TAB   = 1;
local UNLEARNED_TAB = 2;

local eventFrame;

local tradeSkillInfo =
{
  buttons        = {},
  buttonsBg      = nil,
  textures       = {},
  history        = nil,
  historyEnabled = true,
  recipes        = {},
  products       = {},
  resultTexture  = nil,
};

local tradeSkillHistory = setmetatable({},
{
  __index = function(tbl, key)
    tbl[key] = NigglesHistoryClass:New();
    return tbl[key];
  end
});

-------------------------------------------------------------------------------
--             L  O  C  A  L     D  E  F  I  N  I  T  I  O  N  S
-------------------------------------------------------------------------------

local recipeIsPassive;
local tradeSkillInfoUpdate;
local tradeSkillFrameOnShow;
local tradeSkillHistoryBtnEnable;
local tradeSkillHistoryBtnOnClick;
local tradeSkillRecipeShow;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function hooked to 'TradeSkillFrame.Details:RefreshDisplay()'. It
-- shows/hides the textures used to indicate recipe reagents that can be
-- made by the current trade skill and character. It also adds the selected
-- recipe to the history.
--
local function recipeDetailsOnRefresh(self)
  -- Local Variables
  local _;
  local numReagents;
  local reagentBtns = TradeSkillFrame.DetailsFrame.Contents.Reagents;
  local reagentLink;
  local reagentId;
  local recipeId = self.selectedRecipeID;
  local strfind = string.find;
  local textures = tradeSkillInfo.textures;

  -- Check there is a recipe displayed
  numReagents = C_TradeSkillUI.GetRecipeNumReagents(recipeId or 0);
  if ((TradeSkillFrame:IsShown()) and
      (recipeId ~= nil) and
      (numReagents ~= nil)) then
    -- Show/Hide the textures that indicate a reagent can be created
    for idx = 1, numReagents do
      -- Get the reagent's ID
      reagentLink = C_TradeSkillUI.GetRecipeReagentItemLink(recipeId, idx);
      if (reagentLink ~= nil) then
        _, _, reagentId = strfind(reagentLink, "|Hitem:(%d+):");
      else
        reagentId = nil;
      end

      -- Show/Hide the button's texture
      if ((reagentBtns[idx] ~= nil) and
          (reagentBtns[idx]:IsShown()) and
          (textures[idx] ~= nil)) then
        textures[idx]:SetShown(((reagentId ~= nil) and
          (tradeSkillInfo.products[reagentId] ~= nil)));
      end
    end

    -- Add the recipe to the history, if required
    if ((tradeSkillInfo.historyEnabled) and
        (tradeSkillInfo.history ~= nil) and
        (tradeSkillInfo.history:Get() ~= recipeId)) then
      tradeSkillInfo.history:Add(recipeId);
      tradeSkillHistoryBtnEnable();
    end
  else
    -- Hide the textures that indicate a reagent can be created
    for idx = 1, #textures do
      textures[idx]:Hide();
    end
  end

  -- Show/hide the texture that indicates the recipe can be dragged
  tradeSkillInfo.resultTexture:SetShown(
    ((NigglesDBC.settings.ProfGeneralRecipeDrag) and
     (not recipeIsPassive(recipeId))));

  return;
end

--
-- Function to check if a recipe is passive, such as Legion Herbalism and
-- Mining skills
--
recipeIsPassive = function(recipeId)
  -- Local Variables
  local _;
  local castTime;

  -- Get the recipe's cast time
  _, _, _, castTime = GetSpellInfo(recipeId or 0);

  return ((castTime or 0) == 0);
end

--
-- Function called when a recipe reagent is clicked. It checks if the reagent
-- can be created by the current trade skill and is known by the current
-- character. If it is, the recipe is selected.
--
local function recipeReagentBtnOnClick(self, mouseButton)
  -- Local Variables
  local _;
  local reagentId;
  local reagentLink;
  local recipeId = TradeSkillFrame.DetailsFrame.selectedRecipeID;

  -- Check if the button click should be processed
  if ((mouseButton == "LeftButton") and
      (not IsModifierKeyDown()) and
      (recipeId ~= nil)) then
    -- Get a link for the reagent
    reagentLink = C_TradeSkillUI.GetRecipeReagentItemLink(recipeId,
      self.reagentIndex);
    if (reagentLink ~= nil) then
      -- Extract the reagent's ID
      _, _, reagentId = strfind(reagentLink, "|Hitem:(%d+):");

      -- Check if there is a recipe for the reagent
      if (tradeSkillInfo.products[reagentId] ~= nil) then
        -- Show the recipe for the reagent
        tradeSkillRecipeShow(
          bit.band(tradeSkillInfo.products[reagentId],0x00FFFFFF), true);
      end
    end
  end

  return;
end

--
-- Function called when the result icon dragging starts. It picks up the spell
-- for the current recipe, if required
--
local function recipeResultIconOnDragStart(self, mouseButton)
  -- Local Variables
  local recipeId = TradeSkillFrame.DetailsFrame.selectedRecipeID;

  -- Check the feature is enabled and the recipe isn't passive
  if ((NigglesDBC.settings.ProfGeneralRecipeDrag) and
      (not recipeIsPassive(recipeId))) then
    -- Pick up the recipe
    PickupSpell(recipeId);
  end

  return;
end

--
-- Function called when the Trade Skill frame is shown. It enables/disables
-- Niggles features, depending on the current settings.
--
tradeSkillFrameOnShow = function(self)
  -- Local Variables
  local button;
  local idx = 1;
  local reagents = TradeSkillFrame.DetailsFrame.Contents.Reagents
  local texturePath;
  local texture;
  local textures = tradeSkillInfo.textures;

  -- Set up the background texture for the recipe history buttons
  if ((NigglesDBC.settings.ProfGeneralHistory) and
      (tradeSkillInfo.buttonsBg == nil)) then
    texture = TradeSkillFrame.DetailsFrame:CreateTexture(nil, "MEDIUM");
    texture:SetSize(54, 32);
    texture:SetTexture("Interface\\HELPFRAME\\HelpFrameTab-Active");
    texture:SetPoint("BOTTOMLEFT", TradeSkillFrame.DetailsFrame,
      "TOPLEFT", 5, 0);
    tradeSkillInfo.buttonsBg = texture;
  end
  if (tradeSkillInfo.buttonsBg ~= nil) then
    tradeSkillInfo.buttonsBg:SetShown(
      NigglesDBC.settings.ProfGeneralHistory);
  end

  -- Set up the recipe history buttons
  for idx, suffix in ipairs({"PrevPage", "NextPage"}) do
    -- Create the button, if required
    if ((NigglesDBC.settings.ProfGeneralHistory) and
        (tradeSkillInfo.buttons[idx] == nil)) then
      -- Work out the path for the button's textures
      texturePath = "Interface\\Buttons\\UI-SpellbookIcon-"..suffix;

      -- Create the button
      button = CreateFrame("Button", nil, TradeSkillFrame.DetailsFrame);
      button:SetWidth(23);
      button:SetHeight(22);
      button:SetNormalTexture(texturePath.."-Up");
      button:SetPushedTexture(texturePath.."-Down");
      button:SetDisabledTexture(texturePath.."-Disabled");
      button:SetHighlightTexture(
        "Interface\\Buttons\\UI-Common-MouseHilight", "ADD");
      button:SetPoint("BOTTOMLEFT", TradeSkillFrame.DetailsFrame, "TOPLEFT",
        11+((idx-1)*20), 2);
      button:SetID(idx);
      button:SetScript("OnClick", tradeSkillHistoryBtnOnClick);
      button:Disable();
      tradeSkillInfo.buttons[idx] = button;
    else
      button = tradeSkillInfo.buttons[idx];
    end

    -- Show/Hide the button
    if (button ~= nil) then
      button:SetShown(NigglesDBC.settings.ProfGeneralHistory);
    end
  end

  -- Create the textures that indicate creatable recipe reagents, if required
  if (NigglesDBC.settings.ProfGeneralReagents) then
    while (reagents[idx] ~= nil) do
      button = reagents[idx];
      if (textures[idx] == nil) then
        textures[idx] = button:CreateTexture(nil, "BORDER");
        textures[idx]:SetTexture("Interface\\SPELLBOOK\\RotationIconFrame");
        textures[idx]:SetPoint("TOPLEFT", button.Icon, -6, 6);
        textures[idx]:SetPoint("BOTTOMRIGHT", button.Icon, 7, -7);
        textures[idx]:Hide();
      end
      idx = idx+1;
    end
  end

  -- Create the texture that indicates the recipe can be dragged, if required
  if ((NigglesDBC.settings.ProfGeneralReagents) and
      (tradeSkillInfo.resultTexture == nil)) then
    button  = TradeSkillFrame.DetailsFrame.Contents.ResultIcon;
    texture = button:CreateTexture(nil, "OVERLAY");
    texture:SetTexture("Interface\\SPELLBOOK\\RotationIconFrame");
    texture:SetPoint("TOPLEFT", -10, 10);
    texture:SetPoint("BOTTOMRIGHT", 10, -10);
    texture:Show();
    tradeSkillInfo.resultTexture = texture;
  end

  return;
end

--
-- Function to enable/disable the trade skill history buttons
--
tradeSkillHistoryBtnEnable = function()
  -- Enable/disable the history button
  for idx, backward in ipairs({true, false}) do
    if (tradeSkillInfo.buttons[idx] ~= nil) then
      tradeSkillInfo.buttons[idx]:SetEnabled(
        (tradeSkillInfo.history ~= nil) and
        (tradeSkillInfo.history:IsMoveValid(backward)));
    end
  end

  return;
end

--
-- Function called when a Trade skill History button is clicked.
--
tradeSkillHistoryBtnOnClick = function(self, mouseButton)
  -- Local Variables
  local backward = (self:GetID() == 1);

  -- Check the ID of the current trade skill is valid
  if ((tradeSkillInfo.history ~= nil) and
      (tradeSkillInfo.history:IsMoveValid(backward))) then
    tradeSkillRecipeShow(tradeSkillInfo.history:Move(backward), false);
  end

  return;
end

--
-- Function to update the lists of recipes and products of those recipes for
-- the current trade skill
--
tradeSkillInfoUpdate = function()
  -- Local Variables
  local _;
  local currentInfo = {};
  local currentId;
  local numRecipes;
  local productId;
  local productInfo = tradeSkillInfo.products;
  local productLink;
  local strfind = string.find;
  local recipeIds;
  local recipeInfo = tradeSkillInfo.recipes;

  -- Check if the trade skill isn't ready to be interrogated
  if (not C_TradeSkillUI.IsTradeSkillReady()) then
    -- Call the function again in a short while
    C_Timer.After(0.25, tradeSkillInfoUpdate);
    return false;
  end

  -- Get the IDs of all the recipes for the current trade skill
  recipeIds  = C_TradeSkillUI.GetAllRecipeIDs();
  numRecipes = #recipeIds;

  -- Create a list of indices for the trade skill's recipes and their items
  wipe(recipeInfo);
  wipe(productInfo);
  for idx = 1, numRecipes do
    -- Get the item link for the recipe's product
    productLink = C_TradeSkillUI.GetRecipeItemLink(recipeIds[idx]);

    -- Extract the item ID for the product
    _, _, productId = strfind(productLink, "|Hitem:(%d+):");

    -- Save the recipe's product
    recipeInfo[recipeIds[idx]] = (productId or 0);

    -- Save the product's recipe, if the product is an item
    if (productId ~= nil) then
      -- Get info for the recipe
      currentInfo = C_TradeSkillUI.GetRecipeInfo(recipeIds[idx]);

      -- Create a recipe ID, adjusted for rank and learned status
      -- NOTE: If there are ever more than 3 ranks of recipe this code will
      --       need to be modified
      currentId = 0x10000000*
        ((currentInfo.previousRecipeID ~= nil and 2 or 0)+
         (((currentInfo.previousRecipeID ~= nil) ~=
          (currentInfo.nextRecipeID ~= nil)) and 1 or 0));
      if (currentInfo.learned) then
        currentId = 0x10000000+currentId;
      else
        currentId = 0x03000000-currentId;
      end
      currentId = currentId+recipeIds[idx];

      -- Save the product's recipe ID, if higher ranked */
      if ((productInfo[productId] or 0) < currentId) then
        productInfo[productId] = currentId;
      end
    end
  end

  return true;
end

--
-- Function to process events
--
local function tradeSkillOnEvent(self, event, ...)
  -- Process the event
  if (event == "TRADE_SKILL_DATA_SOURCE_CHANGED") then
    -- Get the ID of the trade skill
    local skillId = C_TradeSkillUI.GetTradeSkillLine();
    if (skillId ~= nil) then
      -- Work out which history list to use
      if ((C_TradeSkillUI.IsTradeSkillGuild()) or
          (C_TradeSkillUI.IsTradeSkillLinked()) or
          (C_TradeSkillUI.IsNPCCrafting())) then
        -- Use a special history list for other characters
        tradeSkillHistory[0]:Clear();
        tradeSkillInfo.history = tradeSkillHistory[0];
      else
        tradeSkillInfo.history = tradeSkillHistory[skillId];
      end

      -- Update the trade skill info
      tradeSkillInfoUpdate();
    else
      tradeSkillInfo.history = nil;
    end

    -- Enable/disable the trade skill history buttons
    tradeSkillHistoryBtnEnable();
  elseif (event == "NEW_RECIPE_LEARNED") then
    tradeSkillInfoUpdate();
  elseif (event == "TRADE_SKILL_CLOSE") then
    wipe(tradeSkillInfo.recipes);
    wipe(tradeSkillInfo.products);
  end

  return;
end

--
-- Function to show a specified recipe in the Trade Skill frame. It removes all
-- filtering, un-collapses all categories and makes sure the correct tab is
-- visible. It then selects the recipe and forces it to be visible.
--
tradeSkillRecipeShow = function(recipeId, historyAdd)
  -- Local Variables
  local recipeList = TradeSkillFrame.RecipeList;
  local recipeInfo = C_TradeSkillUI.GetRecipeInfo(recipeId);

  -- Check the recipe's info could be obtained
  if (type(recipeInfo) == "table") then
    -- Close any drop down menus
    CloseDropDownMenus();

    -- Clear all filtering
    C_TradeSkillUI.ClearInventorySlotFilter();
    C_TradeSkillUI.ClearRecipeCategoryFilter();
    C_TradeSkillUI.ClearRecipeSourceTypeFilter();
    C_TradeSkillUI.SetOnlyShowMakeableRecipes(false);
    C_TradeSkillUI.SetOnlyShowSkillUpRecipes(false);
    C_TradeSkillUI.SetRecipeItemNameFilter(nil);
    C_TradeSkillUI.SetRecipeItemLevelFilter(0, 0);
    TradeSkillFrame.SearchBox:SetText("");
    TradeSkillFrame.SearchBox:ClearFocus();

    -- Un-collapse all categories
    recipeList.collapsedCategories = {};

    -- Enable/disable updates to the trade skill history
    tradeSkillInfo.historyEnabled = ((historyAdd == nil) or (historyAdd))

    -- Make sure the correct list is displayed
    if (((recipeInfo.learned) and
         (recipeList.selectedTab ~= LEARNED_TAB)) or
        ((not recipeInfo.learned) and
         (recipeList.selectedTab ~= UNLEARNED_TAB))) then
      -- Change the selected tab
      PanelTemplates_SetTab(recipeList,
        (recipeInfo.learned and LEARNED_TAB or UNLEARNED_TAB));

      -- Update the selected recipes
      C_TradeSkillUI.SetOnlyShowLearnedRecipes(recipeInfo.learned);
      C_TradeSkillUI.SetOnlyShowUnlearnedRecipes(not recipeInfo.learned);
    end

    -- Force a refresh of the list, so the history can be re-enabled
    recipeList:Refresh();
    recipeList:OnUpdate();

    -- Select the correct recipe and make sure it's visible
    recipeList:SelectedAndForceRecipeIDIntoView(recipeInfo.recipeID);

    -- Enable updates to the trade skill history
    tradeSkillInfo.historyEnabled = true;

    -- Update the state of the trade skill history buttons
    tradeSkillHistoryBtnEnable();
  end

  return;
end

-------------------------------------------------------------------------------
--                A  D  D  O  N     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to enable/disable the Niggles trade skill features
--
function L.tradeSkillEnable()
  -- Local Variables
  local button;
  local idx = 1;

  -- Check if the frame used to monitor events has already be created
  if (eventFrame == nil) then
    -- Create the event frame
    eventFrame = CreateFrame("FRAME", nil);
    eventFrame:SetScript("OnEvent", tradeSkillOnEvent);
  end

  -- (Un)register for events, depending on the current settings
  if (NigglesDBC.settings.ProfGeneralReagents) then
    eventFrame:RegisterEvent("TRADE_SKILL_DATA_SOURCE_CHANGED");
    eventFrame:RegisterEvent("TRADE_SKILL_CLOSE");
    eventFrame:RegisterEvent("NEW_RECIPE_LEARNED");
  else
    eventFrame:UnregisterEvent("TRADE_SKILL_DATA_SOURCE_CHANGED");
    eventFrame:UnregisterEvent("TRADE_SKILL_CLOSE");
    eventFrame:UnregisterEvent("NEW_RECIPE_LEARNED");
  end

  -- Update the Trade Skill frame, based on current settings
  if ((TradeSkillFrame ~= nil) and
      (TradeSkillFrame:IsShown())) then
    tradeSkillFrameOnShow(TradeSkillFrame);
  end

  -- Update/Clear the recipe info, based on current settings
  if ((NigglesDBC.settings.ProfGeneralReagents) and
      (C_TradeSkillUI.GetTradeSkillLine() ~= nil)) then
    tradeSkillInfoUpdate();
  else
    -- Clear any trade skill recipe info
    wipe(tradeSkillInfo.recipes);
    wipe(tradeSkillInfo.products);
  end

  return;
end

--
-- Function called when Blizzard's Trade Skill addon is loaded
--
function L.tradeSkillOnLoad()
  -- Local Variables
  local button;
  local idx = 1;
  local reagentBtns = TradeSkillFrame.DetailsFrame.Contents.Reagents;
  local texture;

  -- Hook the Trade Skill Frame
  TradeSkillFrame:HookScript("OnShow", tradeSkillFrameOnShow);

  -- Hook Blizzard functions
  hooksecurefunc(TradeSkillFrame.DetailsFrame, "RefreshDisplay",
    recipeDetailsOnRefresh);

  -- Hook the 'OnClick' script handler for the reagent buttons
  while (reagentBtns[idx] ~= nil) do
    reagentBtns[idx]:HookScript("OnClick", recipeReagentBtnOnClick);
    idx = idx+1;
  end

  -- Hook the 'OnDragStart' script handler for the result icon
  button = TradeSkillFrame.DetailsFrame.Contents.ResultIcon;
  button:RegisterForDrag("LeftButton");
  button:HookScript("OnDragStart", recipeResultIconOnDragStart);

  return;
end
